#!/bin/bash
# Possible Arguments:
# Boot File
# Root File
# Flash Version Prefix (from which to infer the filenames)
# Repository name (either "local" or <ftp_site>)

# Global Variables, Defaulted
PROGRAM="$0"
VERSION="0.9.1"
BOOT_FILE="bootfs.jffs2"
ROOT_FILE="rootfs.jffs2"
TEST_MODE=0
INTERACTIVE=1

# System stuff we need:
LOGGER="/usr/bin/logger"
BASENAME="/usr/bin/basename"
CHKFIRMWARE="/home/httpd/jbmconfig/bin/chkfirmware"
DATE="/bin/date"
NANDWRITE="/sbin/nandwrite"
NANDWRITE_STORAGE="/storage/nandwrite"
NANDTEST="/sbin/nandtest"
BTREBOOTER="/bin/btrebooter"
PREPFLASH="/home/httpd/jbmconfig/bin/prepflash.sh"
FLASH_ERASEALL="/sbin/flash_eraseall"
MOUNT="/bin/mount"
SLEEP="/bin/sleep"
REBOOT="/sbin/reboot"
RM="/bin/rm"
CP="/bin/cp"
LS="/bin/ls"
LN="/bin/ln"
DF="/bin/df"
GREP="/bin/grep"
AWK="/bin/awk"


NEWROOT="/storage/flash"
DEVPATH="$NEWROOT/dev"
BOOTDEVMTD="$DEVPATH/mtd4"
ROOTDEVMTD="$DEVPATH/mtd5"


SENTRY_MODEL=$(modelno 2>&1 | grep "[BT|SN|RAM]\-*9[0-9][0-9][0-9]")

#hopefully this will blink even while the ROM is erased because 
#usleep and the file handle should already be in memory
function blink_sentry_wan()
{
	WAN_LED="/sys/class/leds/wan/brightness"
	local DELAY=500000
	if [ "$1" = "fast" ]; then
		DELAY=100000
	fi

	exec 3>$WAN_LED
	while [ 1 ]; do
		echo 1 >&3
		usleep $DELAY
		echo 0 >&3
		usleep $DELAY
	done
}

function leds_before_update()
{
	local SPEED=$1
	if [ "$SPEED" = "slow" ]; then
		if [ -n "$SENTRY_MODEL" ]; then
			echo 2 > /sys/class/leds/status/blink 2>/dev/null
			( blink_sentry_wan "$SPEED" &> /dev/null ) &
			sleep 1
		else
			echo 4 > /sys/class/leds/pwr/blink 2> /dev/null
			echo 4 > /sys/class/leds/wan/blink 2> /dev/null
		fi
	else
		if [ -n "$SENTRY_MODEL" ]; then
			echo 4 > /sys/class/leds/status/blink 2> /dev/null
			echo 4 > /sys/class/leds/uart/blink 2> /dev/null
		else
			echo 4 > /sys/class/leds/uart/blink 2> /dev/null
			echo 2 > /sys/class/leds/reg/blink 2> /dev/null
		fi
	fi
}

function move_tmp()
{
   #This must be run BEFORE / (root) is made ReadOnly
   #
   #This part should be run only after proccesses (including cell backends)
   #are stopped, since proccesses write to /tmp/
   #But this might not matter if a few files get hidden when the mount is moved

	local ALREADY_MOUNTED=0
	if [ -n "$NEWROOT" ]; then

		NEWTMP="$NEWROOT/tmp"
		NEWDEV="$NEWROOT/dev"

		#check if /tmp/ is already moved, and abort
		/bin/mount  | grep tmpfstmp | grep "/storage" &>/dev/null
		if [ "$?" = "0" ]; then
			echo "/tmp/ Already moved to /storage/"
			ALREADY_MOUNTED=1
		else
			#make sure it is empty
			rm -rf $NEWROOT/*
		fi

		#cp some dev handles for flashing, and console output
		mkdir -p $NEWTMP $NEWDEV
		#cannot copy mtd devices, it copies the content
		mknod $NEWDEV/mtd4    c 90 8
		mknod $NEWDEV/mtd5    c 90 10
		mknod $NEWDEV/null    c 1 3
		mknod $NEWDEV/console c 5 1

		if [ 0 = $ALREADY_MOUNTED ]; then
			mount --move /tmp $NEWTMP
		fi

		#Finally. ln main /tmp/ to the new /tmp
		rm -rf /tmp
		ln -sf $NEWTMP /

	fi
}


for i in ${LOGGER} ${BASENAME} ${NANDWRITE} ${BTREBOOTER} ${CHKFIRMWARE} ${FLASH_ERASEALL} \
         ${MOUNT} ${SLEEP} ${REBOOT} ${RM} ${CP} ${LS} ${LN} ${DF} ${AWK} ${GREP}
do
    if [ ! -x ${i} ]
    then
        # try our best:
        /usr/bin/logger -t ${PROGRAM} "Required command (${i}) is missing or NOT executable!"
        exit 1
    fi
done

# Adjust some stuff, now that we know we have what we need:
PROGRAM=`${BASENAME} $0`
LOGGER="${LOGGER} -t ${PROGRAM}"

function die {
    (${SLEEP} 30; ${REBOOT}) &
    echo " -- FAILURE. Rebooting in 30 seconds..."
    ${LOGGER} " -- FAILURE. Rebooting in 30 seconds..."
    exit 1
}

# Simple "Usage" function to display help for the user:
function Usage {
cat <<EOF
Usage: ${PROGRAM} <options>

Options:
-h                     Display this help information and exit
-f                     Force non-interactive mode (no prompting)
-b <boot_file_name>    The name of the boot file to use for flashing
-r <root_file_name>    The name of the root file to use for flashing
-t                     Validate procedure, but do NOT flash (Test Mode)
-v                     Display version information and exit
EOF
}

# Process incoming arguments:
while getopts "b:dfhr:s:tv" flag
do
    case ${flag} in
        'b')
            if [ -n "${OPTARG}" -a ${OPTARG:0:1} != "-" ]
            then
                BOOT_FILE="${OPTARG}"
            else
                echo "A boot file name MUST be provided with the '-${flag}' argument... Exiting."
                echo ""
                Usage
           exit 0
            fi
        ;;
        'f')
            INTERACTIVE=0
        ;;
        'h')
            Usage
            exit 0
        ;;
        'r')
            if [ -n "${OPTARG}"  -a ${OPTARG:0:1} != "-" ]
            then
                ROOT_FILE="${OPTARG}"
            else
                echo "A root file name MUST be provided with the '-${flag}' argument... Exiting."
                echo ""
                Usage
           exit 0
            fi
        ;;
        't')
            echo "*********** TEST MODE ***********"
            TEST_MODE=1
        ;;
        'v')
            echo "${PROGRAM} Version ${VERSION}"
            exit 0
        ;;
        *)
            echo ""
            Usage
            exit 1
        ;;
    esac
done


##########################################################################
# Confirm everything with the user before continuing (if interactive)...
##########################################################################
echo "VERSION=${VERSION}"
echo "BOOT_FILE=${BOOT_FILE}"
echo "ROOT_FILE=${ROOT_FILE}"

if [ ${INTERACTIVE} -eq 1 ]
then
    echo ""
    echo -n "Continue [y/n]? "
    read answer

    if [ "${answer}" != "y" ]
    then
        exit 0
    fi
fi

${LOGGER} " -- Version ${VERSION} - Started."

#this is for G25 CPU builds, which are not padded
PAD_OPTION=
if [ `grep "AT91SAM9X5" /proc/cpuinfo -c 2> /dev/null` -gt 0 ]; then
   PAD_OPTION="-p"
fi

##########################################################################
# Flash some blinky lights...
##########################################################################
leds_before_update "slow"
	
##########################################################################
# Make sure we have our image files in place
##########################################################################
# Check to make sure the boot file exists
if [ ! -e "${BOOT_FILE}" ]
then
    echo "${BOOT_FILE} does NOT exist. Abort."
    ${LOGGER} " -- ${BOOT_FILE} does NOT exist. Exiting."
    exit 1
fi

# Check to make sure the root file exists
if [ ! -e "${ROOT_FILE}" ]
then
    echo "${ROOT_FILE} does NOT exist. Abort."
    ${LOGGER} " -- ${ROOT_FILE} does NOT exist. Exiting."
    exit 1
fi


##########################################################################
# Move /tmp/ mount point to /storage/flash/tmp or something :
##########################################################################
move_tmp

##########################################################################
# Verify JFFS2 image files:
##########################################################################
echo -n "Verifying file integrity..."

CMD="${CHKFIRMWARE} ${BOOT_FILE}"
${CMD}
if [ $? -ne 0 ]
then
    ${LOGGER} " -- '$CMD' failed: $!"
    die
fi

CMD="${CHKFIRMWARE} ${ROOT_FILE}"
${CMD}
if [ $? -ne 0 ]
then
    ${LOGGER} " -- '$CMD' failed: $!"
    die
fi

echo " Complete."


##########################################################################
# Check to see if we are really gonna do this or not:
##########################################################################
if [ $TEST_MODE -eq 1 ]
then
    echo "This is the point of no return..."
    echo ""
    echo "TMP DIR CONTENTS:"
    ${LS} -al /tmp
    echo ""
    echo "STORAGE DIR CONTENTS:"
    ${LS} -al /storage
    echo ""
    echo "FILESYSTEM STATUS:"
    ${DF} -k
    exit 0
fi

##########################################################################
# Tear down more services that may interrupt the process:
##########################################################################
${PREPFLASH}

##########################################################################
# THIS IS THE POINT OF NO RETURN. RESULTS ARE NO LONGER GUARANTEED!!!
##########################################################################
echo "Beginning flash process with:"
echo "BOOT FILE: ${BOOT_FILE}"
echo "ROOT FILE: ${ROOT_FILE}"
${LOGGER} " -- Beginning flash process with:"
${LOGGER} " -- BOOT FILE: ${BOOT_FILE}"
${LOGGER} " -- ROOT FILE: ${ROOT_FILE}"

##########################################################################
# Flash some blinky lights...
##########################################################################
leds_before_update "fast"

##########################################################################
# Copy binaries over that we'll need after the ROM is gone:
##########################################################################
echo -n "Staging required files..."

CMD="${CP} ${NANDWRITE} ${NANDWRITE_STORAGE}"
${CMD}
if [ $? -ne 0 ]
then
    ${LOGGER} " -- '${CMD}' failed: $!"
    die
fi

##########################################################################
# Copy the libraries upon which the binaries depend:
##########################################################################
if [ ! -L /lib/libc-2.6.so ]
then
    CMD="${CP} /lib/libc-2.6.so /storage/"
    ${CMD}
    if [ $? -ne 0 ]
    then
        ${LOGGER} " -- '${CMD}' failed: $!"
        die
    fi

    CMD="${LN} -sf /storage/libc-2.6.so /lib/libc.so.6"
    ${CMD}
    if [ $? -ne 0 ]
    then
        ${LOGGER} " -- '${CMD}' failed: $!"
        die
    fi
fi

if [ ! -L /lib/ld-2.6.so ]
then
    CMD="${CP} /lib/ld-2.6.so /storage/"
    ${CMD}
    if [ $? -ne 0 ]
    then
        ${LOGGER} " -- '${CMD}' failed: $!"
        die
    fi

    CMD="${LN} -sf /storage/ld-2.6.so /lib/ld-linux.so.3"
    ${CMD}
    if [ $? -ne 0 ]
    then
        ${LOGGER} " -- '${CMD}' failed: $!"
        die
    fi
fi

echo " Complete."

##########################################################################
# Yank the filesystem away and re-write the flash:
##########################################################################
# Re-mount the filesystem as READ-ONLY:
echo -n "Disabling write access..."


sync; sync
CMD="${MOUNT} -n -o remount,ro /"
COUNT=0
RETCODE=1
while [ $COUNT -lt 10 -a $RETCODE -ne 0 ]; do
	COUNT=$(($COUNT+1))
	${CMD}
	RETCODE=$?
	if [ $RETCODE -eq 0 ]; then
		COUNT=10
	else
		echo " -- '${CMD}' failed: $! - attempt $COUNT/10"
		${LOGGER} " -- '${CMD}' failed: $! - attempt $COUNT/10"
		sleep 1
	fi
done

if [ $RETCODE -ne 0 ]
then
    ${LOGGER} " -- '${CMD}' failed: $!"
	#continue on, don't exit.
	#For some unknown reason exiting and rebooting at this point cause file corruption
	#and failure to boot.  Tested 5 times with reount failing and unit flashed good
    #die
fi

echo " Complete."

# Erase the boot partition:
echo "Flashing (will reset in 15 minutes)..."

# Setup a back-door rebooter job:
#Note: 5 mins is too short for big ROMs
(${BTREBOOTER} 900) &

echo "Running nandtest on $BOOTDEVMTD"
echo "`${DATE}` - Running nandtest on $BOOTDEVMTD" >> /storage/testnand.log
${NANDTEST} -m $BOOTDEVMTD | ${AWK} -F "(\r)" '{print $NF}' | ${GREP} -i "bad" >> /storage/testnand.log
${FLASH_ERASEALL} $BOOTDEVMTD

# Replace the boot partition:
${NANDWRITE_STORAGE} ${PAD_OPTION} -a -m $BOOTDEVMTD "${BOOT_FILE}"

# Erase the root partition:
echo "Running nandtest on $ROOTDEVMTD"
echo "`${DATE}` - Running nandtest on $ROOTDEVMTD" >> /storage/testnand.log
${NANDTEST} -m $ROOTDEVMTD | ${AWK} -F "(\r)" '{print $NF}' | ${GREP} -i "bad" >> /storage/testnand.log
${FLASH_ERASEALL} $ROOTDEVMTD

# Replace the root partition:
${NANDWRITE_STORAGE} ${PAD_OPTION} -a -m $ROOTDEVMTD "${ROOT_FILE}"

echo " Complete."

#Reboot
echo "Resetting... "
while [ 1 ]
do
    ${BTREBOOTER}
done

